/*
libfawk - A function-only AWK dialect

Copyright (c) 2020 Tibor 'Igor2' Palinkas. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

  * 1. Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
  * 3. Neither the name of the copyright holder nor the names of its
    contributors may be used to endorse or promote products derived from this
    software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Project page: http://repo.hu/projects/libfawk
Source code: svn://repo.hu/libfawk/trunk
Contact the author: http://igor2.repo.hu/contact.html
*/

#include <stdlib.h>
#include "libfawk.h"
#include "builtin.h"
#include "array.h"
#include "str.h"
#include "symtab.h"
#include "execute.h"
#include "cast.h"
#include <genht/hash.h>
#include "genregex_se.c"

static unsigned fawk_re_ptrhash(const void *k) {
 const unsigned long n = (const unsigned long)k;
 return (n >> 2) ^ (n >> 12);
}
static int fawk_re_ptrkeyeq(const void *a, const void *b) { return a == b; }

static void regexp_str_free_cb(fawk_pkg_t *pkg, fawk_ctx_t *ctx, fawk_str_t *str)
{
	htpp_t *cache = pkg->data[0].ptr;
	fawk_re_t *re = htpp_pop(cache, str);
	if (re != NULL)
		fawk_re_free(re);
}

static void regexp_uninit_cb(fawk_pkg_t *pkg, fawk_ctx_t *ctx)
{
	htpp_t *cache = pkg->data[0].ptr;
	htpp_uninit(cache);
	fawk_free(ctx, cache);
}

static htpp_t *get_cache(fawk_ctx_t *ctx)
{
	fawk_pkg_t *pkg;
	for(pkg = ctx->pkg_head; pkg != NULL; pkg = pkg->next)
		if (pkg->uninit_cb == regexp_uninit_cb)
			return pkg->data[0].ptr;
	return NULL;
}

#define cache_str(S) \
	re = htpp_get(cache, S->data.str); \
	if (re == NULL) htpp_set(cache, S->data.str, re = fawk_re_comp(S->data.str->str));

#define set_arr_str(arr, idx, string, string_len) \
	do { \
		fawk_cell_t *elem = fawk_array_resolve_c(ctx, 1, arr, FAWK_NUM, (fawk_num_t)idx, FAWK_NIL); \
		if (string_len != 0) { elem->type = FAWK_STR; elem->data.str = fawk_str_new_from_literal(ctx, string, string_len); } \
		else elem->type = FAWK_NIL; \
	} while(0)

static void fawk_regexp_match(fawk_ctx_t *ctx, const char *fname, int argc, fawk_cell_t *retval)
{
	htpp_t *cache = get_cache(ctx);
	fawk_re_t *re;
	fawk_cell_t *opat = NULL, *oaddr = NULL;
	int n;

	if ((argc < 2) || (argc > 4) || (FAWK_CFUNC_ARG(1)->type != FAWK_STR))
		return;

	if (fawk_cast_to_str(ctx, FAWK_CFUNC_ARG(0)) != 0)
		return;

	if (argc > 2) {
		opat = FAWK_CFUNC_ARG(2);
		if (opat->type != FAWK_ARRAY)
			return;
	}

	if (argc > 3) {
		oaddr = FAWK_CFUNC_ARG(3);
		if (oaddr->type != FAWK_ARRAY)
			return;
	}

	cache_str(FAWK_CFUNC_ARG(1));

	retval->type = FAWK_NUM;
	retval->data.num = fawk_re_exec(re, FAWK_CFUNC_ARG(0)->data.str->str);

	if (opat != NULL) {
		for(n = 0; n < MAXTAG; n++) {
			if (re->bopat[n] != NULL) {
				set_arr_str(opat, n, re->bopat[n], re->eopat[n]-re->bopat[n]);
			}
		}
	}

	if (oaddr != NULL) {
		for(n = 0; n < MAXTAG; n++) {
			if (re->bopat[n] != NULL) {
				fawk_cell_t *elem = fawk_array_resolve_c(ctx, 1, oaddr, FAWK_NUM, (fawk_num_t)(n*2), FAWK_NIL);
				elem->type = FAWK_NUM; elem->data.num = re->bopat[n] - re->bol;
				elem = fawk_array_resolve_c(ctx, 1, oaddr, FAWK_NUM, (fawk_num_t)(n*2+1), FAWK_NIL);
				elem->type = FAWK_NUM; elem->data.num = re->eopat[n] - re->bopat[n];
			}
		}
	}
}

static void fawk_regexp_sub(fawk_ctx_t *ctx, const char *fname, int argc, fawk_cell_t *retval)
{
	htpp_t *cache = get_cache(ctx);
	fawk_re_t *re;
	fawk_cell_t *str, *v, *arr;
	int first_only;
	char *dst;

	if (argc != 3)
		return;

	if ((fawk_cast_to_str(ctx, FAWK_CFUNC_ARG(0)) != 0) || (fawk_cast_to_str(ctx, FAWK_CFUNC_ARG(1)) != 0))
		return;

	first_only = (fname[0] != 'g');

	cache_str(FAWK_CFUNC_ARG(0));

	v = FAWK_CFUNC_ARG(2);
	if (v->type == FAWK_SYMREF) {
		if (((str = fawk_symtab_deref(ctx, &v->data.symref, 0, &arr)) == NULL) || ((str->type != FAWK_STR) && (str->type != FAWK_STRNUM)))
			return;
		
		retval->type = FAWK_NUM;
		retval->data.num = fawk_re_subst(re, &dst, str->data.str->str, FAWK_CFUNC_ARG(1)->data.str->str, first_only);
		fawk_str_free(ctx, str->data.str);
		str->type = FAWK_STR;
		str->data.str = fawk_str_new_from_literal(ctx, dst, -1);
	}
	else {
		if (fawk_cast_to_str(ctx, v) != 0)
			return;
		fawk_re_subst(re, &dst, v->data.str->str, FAWK_CFUNC_ARG(1)->data.str->str, first_only);
		retval->type = FAWK_STR;
		retval->data.str = fawk_str_new_from_literal(ctx, dst, -1);
	}
}

static void fawk_regexp_split(fawk_ctx_t *ctx, const char *fname, int argc, fawk_cell_t *retval)
{
	htpp_t *cache = get_cache(ctx);
	fawk_re_t *re;
	fawk_cell_t *arr;
	const char *s;
	long n;

	if (argc != 3)
		return;

	if ((fawk_cast_to_str(ctx, FAWK_CFUNC_ARG(0)) != 0) || (FAWK_CFUNC_ARG(1)->type != FAWK_ARRAY) || (fawk_cast_to_str(ctx, FAWK_CFUNC_ARG(2)) != 0))
		return;

	s = FAWK_CFUNC_ARG(0)->data.str->str;
	arr = FAWK_CFUNC_ARG(1);

	if (FAWK_CFUNC_ARG(2)->data.str->str[0] == '\0') { /* empty string: every char is a separate elem */
		for(n = (fname[5] != '0'); *s != '\0'; n++,s++)
			set_arr_str(arr, n, s, 1);
	}
	else { /* split by regex */
		cache_str(FAWK_CFUNC_ARG(2));
		for(n = (fname[5] != '0'); (*s != '\0') && (fawk_re_exec(re, s) > 0); n++) {
			if (re->eopat[0] == re->bol) { re->bopat[0]++; re->eopat[0]++; } /* avoid infinite loop for broken pattern */
			set_arr_str(arr, n, s, re->bopat[0]-re->bol);
			s = re->eopat[0];
		}

		set_arr_str(arr, n, s, -1);
		n++;
	}
	retval->type = FAWK_NUM; retval->data.num = n - (fname[5] != '0');
}

FAWK_API int fawk_regexp_init(fawk_ctx_t *ctx)
{
	fawk_pkg_t *pkg;
	htpp_t *cache;

	pkg = fawk_calloc(ctx, sizeof(fawk_pkg_t), 1);
	if (pkg == NULL) return -1;
	pkg->next = ctx->pkg_head;
	ctx->pkg_head = pkg;

	pkg->str_free_cb = regexp_str_free_cb;
	pkg->uninit_cb = regexp_uninit_cb;
	htpp_init(pkg->data[0].ptr = cache = fawk_malloc(ctx, sizeof(htpp_t)), fawk_re_ptrhash, fawk_re_ptrkeyeq);
	if (cache == NULL) {
		free(pkg);
		return -1;
	}

	fawk_symtab_regcfunc(ctx, "match",  fawk_regexp_match);
	fawk_symtab_regcfunc(ctx, "sub",    fawk_regexp_sub);
	fawk_symtab_regcfunc(ctx, "gsub",   fawk_regexp_sub);
	fawk_symtab_regcfunc(ctx, "split",  fawk_regexp_split);
	fawk_symtab_regcfunc(ctx, "split0", fawk_regexp_split);

	return 0;
}
